"""Tkinter-specific dotplot GUI elements."""

try:
    from Tkinter import Toplevel, Frame, Text, DISABLED, NONE, TOP, YES, BOTH, NORMAL, END
except:
    from _tkinter import Toplevel, Frame, Text, DISABLED, NONE, TOP, YES, BOTH, NORMAL, END

class TkAlignmentFrame(Frame):
    """A frame displaying the ungapped alignment corresponding to the
    cursor position in the dotplot."""
    def __init__(self, parent = None, hseq = "", vseq = "",
                 hpos = 62, vpos = 62, pad = 62, scores = None):
        """Initialize from sequences and starting position.
        If scores is given, it will be used for similarity coloring."""
        Frame.__init__(self, parent)
        self.scores = scores
        self.hseq = hseq
        self.vseq = vseq
        self.hpos = hpos
        self.vpos = vpos
        self.pad = pad

        self.text = Text(self, width = self.pad*2+1, height = 6,
                         state = DISABLED, wrap = NONE)
        self.text.pack(side = TOP, expand = YES, fill = BOTH)

        self.update_text()
        
    def update_text(self, h_id = None):
        if(h_id == None):
            h_id = self.hpos+1
        t = " "*self.pad+str(h_id)+"\n"+" "*self.pad+"|\n"

        aseqs = []
        for (pos, seq) in ((self.hpos, self.hseq), (self.vpos, self.vseq)):
            left = pos - self.pad
            right = pos+self.pad+1
            if(left < 0):
                s = " "*(-left)+seq[:right]
            else:
                s = seq[left:right]
            t += s+"\n"
            aseqs.append(s)

        t += " "*self.pad+"|\n"+" "*self.pad+str(self.vpos+1)+"\n"

        self.text.config(state = NORMAL)
        self.text.delete("1.0",END)
        self.text.insert("1.0", t)
        self.text.tag_delete("I")
        self.text.tag_delete("S")
        for (i, (h,v)) in enumerate(zip(aseqs[0],aseqs[1])):
            try:
                if(h == v):
                    if(h != ' '):
                        self.text.tag_add("I","3.%d" % i,"3.%d" % (i+1))
                        self.text.tag_add("I","4.%d" % i,"4.%d" % (i+1))
                elif((self.scores is not None) and
                     (self.scores[h][v] > 0)):
                    self.text.tag_add("S","3.%d" % i,"3.%d" % (i+1))
                    self.text.tag_add("S","4.%d" % i,"4.%d" % (i+1))
            except KeyError:
                # handle gap comparisons as exceptions
                continue

        self.text.tag_config("I", background = "cyan", foreground = "black")
        self.text.tag_config("S", background = "#56b2de", foreground = "black")
        self.text.config(state = DISABLED)

class TkDotMediator:
    def __init__(self, dotplot, vmin = None, vmax = None):
        import matplotlib.pyplot as plt
        import numpy as np
        
        # Matplotlib (generic)
        (self.fig, self.ax) = plt.subplots(1,1)
        self.dotplot = dotplot
        self.D = np.array(dotplot.d)
        plt.imshow(self.D, interpolation = "none", aspect="auto", cmap = "Greys",
                   extent = [1,len(dotplot.hseq), len(dotplot.vseq), 1],
                   # greyscale
                   vmin = vmin, vmax = vmax)
        self.vpos = self.hpos = 62
        self.vp_line = self.ax.axhline(self.vpos, color = "blue")
        self.hp_line = self.ax.axvline(self.hpos, color = "blue")
        
        # Tk-specific
        win = Toplevel()
        self.alignment = TkAlignmentFrame(win, dotplot.hseq, dotplot.vseq,
                                          vpos = self.vpos, hpos = self.hpos,
                                          scores = dotplot.scores)
        self.alignment.pack()
        
        self.update()
        
        # Disable default plot keybindings
        key_callbacks = self.fig.canvas.callbacks.callbacks[
            "key_press_event"].keys()
        for i in key_callbacks:
            self.fig.canvas.mpl_disconnect(i)
        # Substitute our own keybindings
        self.fig.canvas.mpl_connect("key_press_event", self.keypress)
        # For mouse clicks, we'll do our cross-hairs update in series
        # with the default mouse interactions
        self.fig.canvas.mpl_connect("button_press_event", self.click)

    def update(self):
        self.vp_line.set_ydata([self.vpos, self.vpos])
        self.hp_line.set_xdata([self.hpos, self.hpos])
        self.fig.show()
        
        self.alignment.vpos = self.vpos
        self.alignment.hpos = self.hpos
        self.alignment.update_text()
        
    def click(self, event):
        (x,y) = map(int,(event.xdata, event.ydata))
        if(y < 0):
            self.vpos = 0
        elif(y >= len(self.dotplot.vseq)):
            self.vpos = len(self.dotplot.vseq)-1
        else:
            self.vpos = y
        if(x < 0):
            self.hpos = 0
        elif(x >= len(self.dotplot.hseq)):
            self.hpos = len(self.dotplot.hseq)-1
        else:
            self.hpos = x
        self.update()
        
    def keypress(self, event):
        if(event.key == "enter"):
            self.print_alignment()
        elif(event.key in ("up","k")):
            self.vleft()
        elif(event.key in ("down","j")):
            self.vright()
        elif(event.key in ("left","h")):
            self.hleft()
        elif(event.key in ("right","l")):
            self.hright()
        elif(event.key == "b"):
            self.offleft()
        elif(event.key == "n"):
            self.onright()
        elif(event.key == "y"):
            self.onleft()
        elif(event.key == "u"):
            self.offright()
        
    def vleft(self):
        self.vpos = max(0, self.vpos-1)
        self.update()
    def vright(self):
        self.vpos = min(len(self.dotplot.vseq), self.vpos+1)
        self.update()
    def hleft(self):
        self.hpos = max(0, self.hpos-1)
        self.update()
    def hright(self):
        self.hpos = min(len(self.dotplot.hseq), self.hpos+1)
        self.update()
    def onleft(self):
        self.hpos = max(0, self.hpos-1)
        self.vpos = max(0, self.vpos-1)
        self.update()
    def onright(self):
        self.hpos = min(len(self.dotplot.hseq), self.hpos+1)
        self.vpos = min(len(self.dotplot.vseq), self.vpos+1)
        self.update()
    def offleft(self):
        self.hpos = max(0, self.hpos-1)
        self.vpos = min(len(self.dotplot.vseq), self.vpos+1)
        self.update()
    def offright(self):
        self.hpos = min(len(self.dotplot.hseq), self.hpos+1)
        self.vpos = max(0, self.vpos-1)
        self.update()

    def print_alignment(self):
        print self.alignment.text.get("1.0",END)
