#!/usr/bin/env python
# Time-stamp: <pydotter.py 2017-05-04 10:49:13 Mark Voorhies>
"""
pydotter, a Python clone of Erik Sonnhammer's DOTTER

Based on the algorithms described in Gene 167:GC1 (Sonnhammer and Durbin, 1996)
with heavy borrowing from the contour plot code in EGAD_lib's
InteractionViewer.py (by Arnab Chowdry, Kim Reynolds and Mark Voorhies)
"""

import tempfile, os
import re

from Dotplot import Dotplot, NucleotideDotplot, blosum62, alpha_aa, alpha_nuc
from DotplotTk import TkAlignmentFrame
from FastaFile import FastaFile
try:
    from Tkinter import *
except ImportError:
    from _tkinter import *

class DotterGUI(Frame):
    """An interactive view of a Dotplot."""
    def __init__(self, root, dotplot):
        """Initialize from the result of a dotplot calculation
        (with matrix d, compression T, sequences hseq and vseq,
        and optional reverse-compliment sequence rseq)."""
        Frame.__init__(self, root)
        self.dotplot = dotplot

        self.width = len(dotplot.d[0])
        self.height = len(dotplot.d)
        self.canvas = Canvas(self, width = self.width, height = self.height)
        self.canvas.config(scrollregion = (0,0,self.width,self.height))
        sbar = Scrollbar(self)
        sbar.config(command = self.canvas.yview)
        self.canvas.config(yscrollcommand = sbar.set)
        sbar.pack(side = RIGHT, fill = Y)
        sbar = Scrollbar(self, orient = HORIZONTAL)
        sbar.config(command = self.canvas.xview)
        self.canvas.config(xscrollcommand = sbar.set)
        sbar.pack(side = BOTTOM, fill = X)        
        self.canvas.pack(side = LEFT, expand = YES, fill = BOTH)

        self.img_id = None
        self.cmin = min(min(i) for i in self.dotplot.d)
        self.cmax = max(max(i) for i in self.dotplot.d)
        self.omin = None
        self.omax = None

        self.hpos = self.vpos = self.pad = 62

        self.hmark = None
        self.vmark = None
        self.fmarks = []

        win = Toplevel(root)
        win.title("PyDotter - Gradient Tool")
        self.gradient = SimpleGradient(win, self)
        self.gradient.pack(side = TOP, expand = YES, fill = X)
        
        win = Toplevel(root)
        win.title("PyDotter - Alignment Tool")
        self.alignment = TkAlignmentFrame(win,
                                          hseq = self.dotplot.hseq,
                                          vseq = self.dotplot.vseq,
                                          scores = self.dotplot.scores)
        self.alignment.pack()

        if(self.dotplot.rseq is not None):
            self.revcomp = TkAlignmentFrame(win,
                                            hseq = self.dotplot.rseq,
                                            vseq = self.dotplot.vseq,
                                            scores = self.dotplot.scores)
            self.revcomp.pack()
        else:
            self.revcomp = None

        # DOTTER default
        self.set_gradient(40,100)

        for widget in (self,win):
            # arrow-key directions
            widget.bind("<Up>", self.vleft)
            widget.bind("<Down>", self.vright)
            widget.bind("<Left>", self.hleft)
            widget.bind("<Right>", self.hright)
            # netack/vi directions
            widget.bind("k", self.vleft)
            widget.bind("j", self.vright)
            widget.bind("h", self.hleft)
            widget.bind("l", self.hright)
            widget.bind("b", self.offleft)
            widget.bind("u", self.offright)
            widget.bind("y", self.onleft)
            widget.bind("n", self.onright)
            widget.bind("<Return>", self.print_alignment)
            widget.focus()

        self.canvas.bind("<Button-1>", self.click)

    def click(self, event):
        y = int(self.canvas.canvasy(event.y))*self.dotplot.T
        if(y < 0):
            self.vpos = 0
        elif(y >= len(self.dotplot.vseq)):
            self.vpos = len(self.dotplot.vseq)-1
        else:
            self.vpos = y
        x = int(self.canvas.canvasx(event.x))*self.dotplot.T
        if(x < 0):
            self.hpos = 0
        elif(x >= len(self.dotplot.hseq)):
            self.hpos = len(self.dotplot.hseq)-1
        else:
            self.hpos = x
        self.update()

    def vleft(self, event):
        self.vpos = max(0, self.vpos-1)
        self.update()
    def vright(self, event):
        self.vpos = min(len(self.dotplot.vseq), self.vpos+1)
        self.update()
    def hleft(self, event):
        self.hpos = max(0, self.hpos-1)
        self.update()
    def hright(self, event):
        self.hpos = min(len(self.dotplot.hseq), self.hpos+1)
        self.update()
    def onleft(self, event):
        self.hpos = max(0, self.hpos-1)
        self.vpos = max(0, self.vpos-1)
        self.update()
    def onright(self, event):
        self.hpos = min(len(self.dotplot.hseq), self.hpos+1)
        self.vpos = min(len(self.dotplot.vseq), self.vpos+1)
        self.update()
    def offleft(self, event):
        self.hpos = max(0, self.hpos-1)
        self.vpos = min(len(self.dotplot.vseq), self.vpos+1)
        self.update()
    def offright(self, event):
        self.hpos = min(len(self.dotplot.hseq), self.hpos+1)
        self.vpos = max(0, self.vpos-1)
        self.update()

    def print_alignment(self, event):
        print self.alignment.text.get("1.0",END)

    def update(self):
        hpix = self.hpos/self.dotplot.T
        vpix = self.vpos/self.dotplot.T

        for i in self.fmarks:
            self.canvas.delete(i)
        for i in self.dotplot.hfeatures:
            L = i.Locus()
            self.fmarks.append(self.canvas.create_line(
                L.start/self.dotplot.T,0,
                L.start/self.dotplot.T,self.height,
                fill = "purple"))
            self.fmarks.append(self.canvas.create_line(
                L.stop/self.dotplot.T,0,
                L.stop/self.dotplot.T,self.height,
                fill = "purple"))
        for i in self.dotplot.vfeatures:
            L = i.Locus()
            self.fmarks.append(self.canvas.create_line(
                0,L.start/self.dotplot.T,
                self.width,L.start/self.dotplot.T,
                fill = "purple"))
            self.fmarks.append(self.canvas.create_line(
                0,L.stop/self.dotplot.T,
                self.width,L.stop/self.dotplot.T,
                fill = "purple"))

        if(self.hmark is not None):
            self.canvas.delete(self.hmark)
        self.hmark = self.canvas.create_line(hpix,0,hpix,self.height,
                                             fill = "blue")
        if(self.vmark is not None):
            self.canvas.delete(self.vmark)
        self.vmark = self.canvas.create_line(0,vpix,self.width,vpix,
                                             fill = "blue")

        self.alignment.hpos = self.hpos
        self.alignment.vpos = self.vpos
        self.alignment.update_text()

        if(self.revcomp):
            self.revcomp.hpos = len(self.dotplot.rseq)-self.hpos-1
            self.revcomp.vpos = self.vpos
            self.revcomp.update_text(h_id = self.hpos+1)

    def set_gradient(self, gmin, gmax):
        if((gmin == self.omin) and (gmax == self.omax)):
            return
        (self.omin, self.omax) = (gmin, gmax)
        self.draw(gmin, gmax)

    def draw(self, cmin, cmax):
        # TODO: we require one write to a temp file because PhotoImage
        #       can not read PGM directly from memory.  It _should_
        #       be possible to do all updates in memory via PhotoImage.put.
        temp_file = tempfile.mkstemp()
        self.write_pgm(self.dotplot.d, temp_file[0],
                       cmin = (self.cmax-self.cmin)*float(cmin)/255+self.cmin,
                       cmax = (self.cmax-self.cmin)*float(cmax)/255+self.cmin,
                       gmin = 255, gmax = 0)
        os.close(temp_file[0])
        self.image = PhotoImage(file=temp_file[1])
        os.remove(temp_file[1])
        if(self.img_id is not None):
            self.canvas.delete(self.img_id)
        self.img_id = self.canvas.create_image(
            (0,0),anchor=NW,image= self.image)

        self.update()

    def write_pgm(self, d, fd, cmin, cmax, gmin, gmax):
        print "Drawing cmin: %s cmax: %s gmin: %s gmax: %s\n" % (
            cmin,cmax,gmin,gmax)

        if(cmin == cmax):
            cmax += 1

        scale = float(gmax-gmin)/(cmax - cmin)
        offset = gmin - scale*cmin

        os.write(fd,"P5 %d %d %d\n" % (len(d[0]),len(d),max(gmin,gmax)))
        for i in d:
            for j in i:
                if(j <= cmin):
                    v = gmin
                elif(j >= cmax):
                    v = gmax
                else:
                    v = scale*j+offset
                os.write(fd,"%c" % chr(int(v)))

        print "written"

class SimpleGradient(Frame):
    """Simple non-modal dialog for controlling the dotplot grayscale."""
    # TODO: Given faster rendering of the dotplot, switch this for a
    #       more interactive gradient slider (as in DOTTER).
    def __init__(self, parent, plot):
        Frame.__init__(self, parent)
        self.plot = plot
        self.min = StringVar()
        self.min.set("40")
        self.max = StringVar()
        self.max.set("100")
        e = Entry(self, textvariable = self.min)
        e.pack()
        e.bind("<Return>", self.update_min)
        e = Entry(self, textvariable = self.max)
        e.pack()
        e.bind("<Return>", self.update_max)
    
    def update_min(self, event):
        try:
            cmin = int(self.min.get())
        except:
            cmin = 0
            self.min.set(cmin)
        if(cmin < 0):
            cmin = 0
            self.min.set(cmin)
        if(cmin > 254):
            cmin = 254
            self.min.set(cmin)
        cmax = int(self.max.get())
        if(cmax < cmin):
            cmax = cmin+1
            self.max.set(cmax)
        self.plot.set_gradient(cmin,cmax)

    def update_max(self, event):
        cmin = int(self.min.get())
        try:
            cmax = int(self.max.get())
        except:
            cmax = 255
            self.max.set(cmax)
        if(cmax < 1):
            cmax = 1
            self.max.set(cmax)
        if(cmax > 255):
            cmax = 255
            self.max.set(cmax)
        if(cmax < cmin):
            cmin = cmax-1
            self.min.set(cmin)
        self.plot.set_gradient(cmin,cmax)
    

if(__name__ == "__main__"):
    import sys
    from optparse import OptionParser
    parser = OptionParser(
        usage = "usage: %prog [options] hseq.fasta vseq.fasta")

    parser.add_option(
        "-W", "--window", dest = "window",
        help = "window size",
        type = "int",
        default = 25)

    (options, args) = parser.parse_args()

    if(len(args) < 2):
        parser.print_help(sys.stderr)
        sys.exit(1)

    if(options.window < 1):
        sys.stderr.write("Window size must be positive\n")
        sys.exit(1)

    hseq = FastaFile(open(args[0])).seqs.values()[0].upper()

    vseq = FastaFile(open(args[1])).seqs.values()[0].upper()

    hfeatures = []
    vfeatures = []

    if(re.search("^[ATUGCN]+$", hseq+vseq)):
        print "Calculating nucleotide dotplot..."
        dotplot = NucleotideDotplot(hseq, vseq, options.window,
                                    hfeatures = hfeatures,
                                    vfeatures = vfeatures)
    else:
        print "Calculating protein dotplot..."
        dotplot = Dotplot(hseq, vseq, alpha_aa, blosum62, options.window,
                          hfeatures = hfeatures,
                          vfeatures = vfeatures)

    print "Rendering..."
    
    root = Tk()
    gui = DotterGUI(root, dotplot)
    gui.pack(side = TOP, expand = YES, fill = BOTH)
    root.title("PyDotter")
    root.mainloop()
